from __future__ import absolute_import

import json
from urllib.parse import urlencode

import tornado.web
from tornado.web import MissingArgumentError
from tornado.httpclient import AsyncHTTPClient
from tornado import gen

from sqlalchemy import column, bindparam, and_, func, select
from sqlalchemy.dialects import postgresql

from .encoders import DateTimeAwareJSONEncoder, DateTimeAwareJSONDecoder


class BaseAPIHandler(tornado.web.RequestHandler):

    def data_received(self, chunk):
        pass

    def initialize(self, psql):
        self.set_header("Content-Type", "application/json; charset=utf-8")
        self.psql = psql

    def get_arg(self, value, name, clazz, default):
        if value:
            return clazz(value)
        else:
            if default != '__NONE__':
                return default
            raise MissingArgumentError(name)

    def get_int(self, value, name, default='__NONE__'):
        return self.get_arg(value, name, int, default)

    def prepare(self):
        if "Content-Type" in self.request.headers and \
                self.request.headers["Content-Type"].startswith("application/json"):
            self.json_args = json.loads(self.request.body.decode(), cls=DateTimeAwareJSONDecoder)
        else:
            self.json_args = None

    @gen.coroutine
    def _execute_query(self, alchemy_query):
        query = alchemy_query.compile(dialect=postgresql.dialect())
        cursor = yield self.psql.execute(str(query), query.params)
        return cursor


class SingleRESTAPIHandler(BaseAPIHandler):

    def get_from(self):
        return self.table

    def get_query(self, *args):
        id = self.get_int(args[0], 'id')
        return self.get_from().select().where(self.table.c.id == id)

    @gen.coroutine
    def get_object_dict(self, *args):
        query = self.get_query(*args).compile(dialect=postgresql.dialect())
        cursor = yield self.psql.execute(str(query), query.params)
        row = cursor.fetchone()
        raise gen.Return(row)

    @gen.coroutine
    def get(self, *args):
        row = yield self.get_object_dict(*args)
        if row is None:
            self.set_status(404)
            return
        self.write(json.dumps(row, cls=DateTimeAwareJSONEncoder))
        self.set_status(200)

    @gen.coroutine
    def options(self, *args):
        self.set_status(200)

    @gen.coroutine
    def put_object_dict(self, id, params):
        update_query = self.table.update().where(self.table.c.id == id).values(
            **params
        ).compile(dialect=postgresql.dialect())
        yield self.psql.execute(str(update_query), update_query.params)

    @gen.coroutine
    def put(self, *args):
        cursor = yield self._execute_query(self.get_query(*args))
        row = cursor.fetchone()
        if row is None:
            self.set_status(404)
            return
        yield self.put_object_dict(row["id"], self.json_args)
        yield self.get(*args)

    @gen.coroutine
    def delete(self, *args):
        cursor = yield self._execute_query(self.get_query(*args))
        row = cursor.fetchone()
        if row is None:
            self.set_status(404)
            return
        delete_query = self.table.delete().where(self.table.c.id == row["id"]).compile(dialect=postgresql.dialect())
        yield self.psql.execute(str(delete_query), delete_query.params)
        self.set_status(204)


class ProxyAPIHandler(BaseAPIHandler):
    @gen.coroutine
    def get(self):
        client = AsyncHTTPClient()
        res = yield client.fetch(self.url + '?' + urlencode({k: v[0] for k, v in self.request.arguments.items()}))
        self.write(res.body)


class ListRESTAPIHandler(BaseAPIHandler):

    def get_from(self):
        return self.table

    def get_query(self, **filters):
        query = self.get_from().select()
        for k, v in filters.items():
            field, _, filter_ = k.partition('__')
            if not filter_:
                query = query.where(column(field) == v)
            elif filter_ == 'startswith':
                query = query.where(getattr(self.table.c, field).startswith(v))
            elif filter_ == 'contains':
                query = query.where(getattr(self.table.c, field).contains(v))
            elif filter_ == 'icontains':
                query = query.where(getattr(self.table.c, field).ilike('%' + v + '%'))
            elif filter_ == 'any':
                query = query.where(getattr(self.table.c, field).any(v))
            elif filter_ == 'ne':
                query = query.where(getattr(self.table.c, field) != v)
        return query

    @gen.coroutine
    def get_object_list(self, query):
        cursor = yield self._execute_query(query)
        raise gen.Return(cursor.fetchall())

    @gen.coroutine
    def get_count(self, query):
        count_cursor = yield self._execute_query(select([func.count()]).select_from(query.alias('items')))
        raise gen.Return(count_cursor.fetchone()["count_1"])

    def get_sort_clause(self, sort_field, sort_dir):
        sort_clause = column(sort_field.decode())
        if sort_dir.decode() == 'DESC':
            sort_clause = sort_clause.desc()
        else:
            sort_clause = sort_clause.asc()
        return sort_clause

    def serialize(self, rows):
        return json.dumps(rows, cls=DateTimeAwareJSONEncoder)

    @gen.coroutine
    def get(self, *args):
        page = self.request.arguments.pop('_page', [None])[0]
        per_page = self.request.arguments.pop('_perPage', [None])[0]
        sort_field = self.request.arguments.pop('_sortField', [None])[0]
        sort_dir = self.request.arguments.pop('_sortDir', [None])[0]
        filters = json.loads(self.request.arguments.pop('_filters', [b'{}'])[0].decode())
        arguments = {k: v[0].decode('utf-8') if v else v for k, v in self.request.arguments.items()}

        query = self.get_query(**filters)
        item_count = yield self.get_count(query)

        if page:
            query = query.limit(int(per_page)).offset(int(per_page) * (int(page) - 1))
        if sort_field:
            sort_clause = self.get_sort_clause(sort_field, sort_dir)
            query = query.order_by(sort_clause)

        rows = yield self.get_object_list(query)

        self.set_header('X-Total-Count', item_count)
        self.write(self.serialize(rows))
        self.set_status(200)

    @gen.coroutine
    def options(self):
        self.set_status(200)

    @gen.coroutine
    def post_object_dict(self, params):
        args = {k: v for k, v in params.items() if hasattr(self.table.c, k)}
        query = self.table.insert(returning=[self.table.c.id]).values(**args)
        cursor = yield self._execute_query(query)
        raise gen.Return(cursor.fetchone()['id'])

    @gen.coroutine
    def post(self, *args):
        yield self.post_object_dict(self.json_args)
        self.set_status(201)



from __future__ import absolute_import

from datetime import datetime, timedelta
from json import JSONEncoder, JSONDecoder

class TimeDeltaJSONEncoder(JSONEncoder):
    """
    Converts a python object, where datetime and timedelta objects are converted
    into objects that can be decoded using the DateTimeAwareJSONDecoder.

    Adapted from `gist <https://gist.github.com/majgis/4200488>`_
    """

    def default(self, obj):
        if isinstance(obj, timedelta):
            return {
                'seconds': obj.total_seconds(),
                }

        else:
            return JSONEncoder.default(self, obj)



class DateTimeAwareJSONEncoder(JSONEncoder):
    """
    Converts a python object, where datetime and timedelta objects are converted
    into objects that can be decoded using the DateTimeAwareJSONDecoder.

    Adapted from `gist <https://gist.github.com/majgis/4200488>`_
    """

    def default(self, obj):
        if isinstance(obj, datetime):
            return {
                '__type__': 'datetime',
                'year': obj.year,
                'month': obj.month,
                'day': obj.day,
                'hour': obj.hour,
                'minute': obj.minute,
                'second': obj.second,
                'microsecond': obj.microsecond,
                }

        elif isinstance(obj, timedelta):
            return {
                '__type__' : 'timedelta',
                'days': obj.days,
                'hours': obj.seconds // 3600,
                'minutes': (obj.seconds // 60) % 60,
                'seconds': obj.seconds,
                'microseconds': obj.microseconds,
                }

        else:
            return JSONEncoder.default(self, obj)


class DateTimeAwareJSONDecoder(JSONDecoder):
    """
    Converts a json string, where datetime and timedelta objects were converted
    into objects using the DateTimeAwareJSONEncoder, back into a python object.
    """

    def __init__(self, **kwargs):
        JSONDecoder.__init__(self, object_hook=self.dict_to_object, **kwargs)

    def dict_to_object(self, d):
        if '__type__' not in d:
            return d

        type = d.pop('__type__')
        if type == 'datetime':
            return datetime(**d)
        elif type == 'timedelta':
            return timedelta(**d)
        else:
            # Oops... better put this back together.
            d['__type__'] = type
            return d




from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='tornado-alchemy-rest',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.1.2',

    description='Tornado REST API',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/EzyInsights/tornado-alchemy-rest',

    # Author details
    author='EzyInsights team',
    author_email='andrii.kostenko@kampiki.com',

    # Choose your license
    license='Apache',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',

        'Intended Audience :: Developers',

        'License :: OSI Approved :: Apache Software License',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    # What does your project relate to?
    keywords='tornado sqlalchemy alchemy rest api postgresql',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['tornado', 'sqlalchemy'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': ['check-manifest'],
        'test': ['coverage'],
    },
)
